import math
import logging
from os import path
from typing import List, Union

from keras import layers, models
import numpy as np
import cv2

from .. import core, utils
from .detector import Detector

log = logging.getLogger(__name__)

WEIGHTS_CONFIG = {
    "pnet": {
        "hash":
        "3285cf7a3de2651c5784cb9e32013f5919aae95fd1ed1bc371dd9691affb39af",  # noqa: E501
        "url":
        "https://storage.googleapis.com/miradata/weights/mtcnn/det1.npy",
    },
    "rnet": {
        "hash":
        "716b8b83e42476791c9096f14dbb09fefc88bf5c7ec876b1683f9acd52b3f39c",  # noqa: E501
        "url":
        "https://storage.googleapis.com/miradata/weights/mtcnn/det2.npy",
    },
    "onet": {
        "hash":
        "396ead803d85d3443307ff8f45fb6aed2536579b415a4f4d4cb8f93ea6b1476a",  # noqa: E501
        "url":
        "https://storage.googleapis.com/miradata/weights/mtcnn/det3.npy",
    },
}


def make_pnet(input_shape=(12, 12, 3)):
    x = layers.Input(input_shape)
    conv1 = layers.Conv2D(kernel_size=(3, 3),
                          filters=10,
                          strides=(1, 1),
                          padding="valid",
                          name="conv1")(x)
    relu1 = layers.PReLU(name="PReLU1", shared_axes=[1, 2])(conv1)
    mp1 = layers.MaxPooling2D(pool_size=(2, 2),
                              strides=(2, 2),
                              padding="valid",
                              name="pool1")(relu1)
    conv2 = layers.Conv2D(kernel_size=(3, 3),
                          filters=16,
                          strides=(1, 1),
                          padding="valid",
                          name="conv2")(mp1)
    relu2 = layers.PReLU(name="PReLU2", shared_axes=[1, 2])(conv2)
    conv3 = layers.Conv2D(kernel_size=(3, 3),
                          filters=32,
                          strides=(1, 1),
                          padding="valid",
                          name="conv3")(relu2)
    relu3 = layers.PReLU(name="PReLU3", shared_axes=[1, 2])(conv3)
    conv4_1 = layers.Conv2D(kernel_size=(1, 1),
                            filters=2,
                            strides=(1, 1),
                            padding="valid",
                            name="conv4-1")(relu3)
    prob1 = layers.Softmax(axis=3, name="prob1")(conv4_1)
    conv4_2 = layers.Conv2D(kernel_size=(1, 1),
                            filters=4,
                            strides=(1, 1),
                            name="conv4-2")(relu3)
    return models.Model(inputs=x, outputs=[prob1, conv4_2])


def make_rnet(input_shape=(24, 24, 3)):
    x = layers.Input(input_shape)
    conv1 = layers.Conv2D(kernel_size=(3, 3),
                          filters=28,
                          strides=(1, 1),
                          padding="valid",
                          name="conv1")(x)
    relu1 = layers.PReLU(name="prelu1", shared_axes=[1, 2])(conv1)
    mp1 = layers.MaxPooling2D(pool_size=(3, 3),
                              strides=(2, 2),
                              padding="same",
                              name="pool1")(relu1)
    conv2 = layers.Conv2D(kernel_size=(3, 3),
                          filters=48,
                          strides=(1, 1),
                          padding="valid",
                          name="conv2")(mp1)
    relu2 = layers.PReLU(name="prelu2", shared_axes=[1, 2])(conv2)
    mp2 = layers.MaxPooling2D(pool_size=(3, 3),
                              strides=(2, 2),
                              padding="valid",
                              name="pool2")(relu2)
    conv3 = layers.Conv2D(kernel_size=(2, 2),
                          filters=64,
                          strides=(1, 1),
                          padding="valid",
                          name="conv3")(mp2)
    relu3 = layers.PReLU(name="prelu3", shared_axes=[1, 2])(conv3)
    conv4 = layers.Conv2D(kernel_size=(3, 3),
                          filters=128,
                          strides=(1, 1),
                          padding="valid",
                          name="conv4")(relu3)
    relu4 = layers.PReLU(name="prelu4", shared_axes=[1, 2])(conv4)
    conv5_1 = layers.Conv2D(kernel_size=(1, 1),
                            filters=2,
                            strides=(1, 1),
                            padding="valid",
                            name="conv5-1")(relu4)
    prob1 = layers.Softmax(axis=3)(conv5_1)
    conv5_2 = layers.Conv2D(kernel_size=(1, 1),
                            filters=4,
                            strides=(1, 1),
                            padding="valid",
                            name="conv5-2")(relu4)
    return models.Model(inputs=x, outputs=[prob1, conv5_2])


def make_onet(input_shape=(48, 48, 3)):
    x = layers.Input(input_shape)
    conv1 = layers.Conv2D(kernel_size=(3, 3),
                          filters=32,
                          strides=(1, 1),
                          padding="valid",
                          name="conv1")(x)
    relu1 = layers.PReLU(name="prelu1", shared_axes=[1, 2])(conv1)
    mp1 = layers.MaxPooling2D(pool_size=(3, 3),
                              strides=(2, 2),
                              padding="same",
                              name="pool1")(relu1)
    conv2 = layers.Conv2D(kernel_size=(3, 3),
                          filters=64,
                          strides=(1, 1),
                          padding="valid",
                          name="conv2")(mp1)
    relu2 = layers.PReLU(name="prelu2", shared_axes=[1, 2])(conv2)
    mp2 = layers.MaxPooling2D(pool_size=(3, 3),
                              strides=(2, 2),
                              padding="valid",
                              name="pool2")(relu2)
    conv3 = layers.Conv2D(kernel_size=(3, 3),
                          filters=64,
                          strides=(1, 1),
                          padding="valid",
                          name="conv3")(mp2)
    relu3 = layers.PReLU(name="prelu3", shared_axes=[1, 2])(conv3)
    mp3 = layers.MaxPooling2D(pool_size=(2, 2),
                              strides=(2, 2),
                              padding="same",
                              name="pool3")(relu3)
    conv4 = layers.Conv2D(kernel_size=(2, 2),
                          filters=128,
                          strides=(1, 1),
                          padding="valid",
                          name="conv4")(mp3)
    relu4 = layers.PReLU(name="prelu4", shared_axes=[1, 2])(conv4)
    conv5 = layers.Conv2D(kernel_size=(3, 3),
                          filters=256,
                          padding="valid",
                          name="conv5")(relu4)
    relu5 = layers.PReLU(name="prelu5", shared_axes=[1, 2])(conv5)
    conv6_1 = layers.Conv2D(kernel_size=(1, 1), filters=2,
                            name="conv6-1")(relu5)
    prob1 = layers.Softmax(axis=3, name="prob1")(conv6_1)
    conv6_2 = layers.Conv2D(kernel_size=(1, 1), filters=4,
                            name="conv6-2")(relu5)
    conv6_3 = layers.Conv2D(kernel_size=(1, 1), filters=10,
                            name="conv6-3")(relu5)
    return models.Model(inputs=x, outputs=[prob1, conv6_2, conv6_3])


def load_weights(network, weights_file):
    for layer_name, weights in np.load(weights_file,
                                       encoding="latin1").item().items():
        layer = network.get_layer(layer_name)
        if layer_name.lower().startswith("conv"):
            coefs = weights["weights"].reshape(
                layer.weights[0].shape.as_list())
            biases = weights["biases"].reshape(
                layer.weights[1].shape.as_list())
            layer.set_weights(weights=[coefs, biases])
        elif layer_name.lower().startswith("prelu"):
            alpha = weights["alpha"].reshape(layer.weights[0].shape)
            layer.set_weights(weights=[alpha])
        else:
            raise NotImplementedError("Cannot handle layer: " + layer_name)


def nms(boxes, threshold, method):
    if boxes.size == 0:
        return np.empty((0, 3))
    x1 = boxes[:, 0]
    y1 = boxes[:, 1]
    x2 = boxes[:, 2]
    y2 = boxes[:, 3]
    s = boxes[:, 4]
    area = (x2 - x1 + 1) * (y2 - y1 + 1)
    sortIdx = np.argsort(s)
    pick = np.zeros_like(s, dtype=np.int16)
    counter = 0
    while sortIdx.size > 0:
        i = sortIdx[-1]
        pick[counter] = i
        counter += 1
        idx = sortIdx[0:-1]
        xx1 = np.maximum(x1[i], x1[idx])
        yy1 = np.maximum(y1[i], y1[idx])
        xx2 = np.minimum(x2[i], x2[idx])
        yy2 = np.minimum(y2[i], y2[idx])
        w = np.maximum(0.0, xx2 - xx1 + 1)
        h = np.maximum(0.0, yy2 - yy1 + 1)
        inter = w * h
        if method == "Min":
            o = inter / np.minimum(area[i], area[idx])
        else:
            o = inter / (area[i] + area[idx] - inter)
        sortIdx = sortIdx[np.where(o <= threshold)]
    pick = pick[0:counter]
    return pick


def calculate_boxes(clf,
                    reg,
                    scale,
                    threshold,
                    cell_size,
                    stride=2,
                    offset=(0, 0)):
    """Use heatmap to generate bounding boxes"""
    y, x = np.where(clf[..., 1].transpose() >= threshold)
    score = clf[x, y, 1:]
    reg = reg[x, y]
    bb = np.vstack([y, x]).transpose()
    x1y1 = np.fix((stride * bb + 1) / scale) + offset
    x2y2 = np.fix((stride * bb + 1 + cell_size - 1) / scale) + offset
    x1y1x2y2sd = np.concatenate([x1y1, x2y2, score, reg], axis=1)
    return x1y1x2y2sd


def pad_and_square(x1y1x2y2s, width, height, pad=0.1):
    xpad = pad * (x1y1x2y2s[:, 2] - x1y1x2y2s[:, 0])
    ypad = pad * (x1y1x2y2s[:, 3] - x1y1x2y2s[:, 1])
    x1y1x2y2s[:, 0] -= xpad
    x1y1x2y2s[:, 2] += xpad
    x1y1x2y2s[:, 1] -= ypad
    x1y1x2y2s[:, 3] += ypad
    x1y1x2y2s[:, 0:4:2] = x1y1x2y2s[:, 0:4:2].clip(0, width)
    x1y1x2y2s[:, 1:4:2] = x1y1x2y2s[:, 1:4:2].clip(0, height)
    length = (x1y1x2y2s[:, 2:4] - x1y1x2y2s[:, :2]).max(axis=1)
    x1y1x2y2s[:, 2:4] = x1y1x2y2s[:, :2] + length[:, np.newaxis]


def apply_regression(boxes):
    w = boxes[:, 2] - boxes[:, 0]
    h = boxes[:, 3] - boxes[:, 1]
    boxes[:, 0] += boxes[:, 5] * w
    boxes[:, 1] += boxes[:, 6] * h
    boxes[:, 2] += boxes[:, 7] * w
    boxes[:, 3] += boxes[:, 8] * h
    boxes[:, 5:] = 0
    w = boxes[:, 2] - boxes[:, 0]
    h = boxes[:, 3] - boxes[:, 1]


def run_stage(network, image, x1y1x2y2s, cell_size, threshold):
    crops = [
        image[y1:y2, x1:x2]
        for x1, y1, x2, y2 in x1y1x2y2s[:, :4].round(0).astype("int32")
    ]
    crops = [
        image for image in crops if image.shape[0] > 0 and image.shape[1] > 0
    ]
    X, scales = zip(*[image.fit(cell_size, cell_size) for image in crops])
    X = np.float32([x.transpose(1, 0, 2).scaled(-1, 1) for x in X])
    y = [v.transpose(0, 2, 1, 3) for v in network.predict(X)]
    x1y1x2y2s = np.concatenate([
        calculate_boxes(
            clf,
            box,
            scale=scale,
            threshold=threshold,
            offset=(current[0], current[1]),
            cell_size=cell_size,
        ) for clf, box, scale, current in zip(*y[:2], scales, x1y1x2y2s)
    ])
    apply_regression(x1y1x2y2s)
    return x1y1x2y2s


class MTCNN(Detector):
    """A detector wrapping MTCNN for face detection. Based on the
    implementation found as part of
    `facenet <https://github.com/davidsandberg/facenet>`_. The only
    difference is we've implemented it using keras. Training is not
    supported at this time. The model is documented in `the original
    paper
    <https://kpzhang93.github.io/MTCNN_face_detection_alignment/paper/spl.pdf>`_.

    Attributes:
        pnet: The pnet model (stage 1)
        rnet: The rnet model (stage 2)
        onet: The onet model (stage 3)
    """
    def __init__(self):
        input_shape = (None, None, 3)
        self.annotation_config = core.AnnotationConfiguration(["face"])
        self.pnet = make_pnet(input_shape=input_shape)
        self.rnet = make_rnet(input_shape=input_shape)
        self.onet = make_onet(input_shape=input_shape)

        load_weights(
            self.pnet,
            utils.get_file(
                origin=WEIGHTS_CONFIG["pnet"]["url"],  # noqa; E501
                file_hash=WEIGHTS_CONFIG["pnet"]["hash"],  # noqa; E501
                cache_subdir=path.join("weights", "mtcnn"),
                hash_algorithm="sha256",
                extract=False,
            ),
        )
        load_weights(
            self.rnet,
            utils.get_file(
                origin=WEIGHTS_CONFIG["rnet"]["url"],  # noqa; E501
                file_hash=WEIGHTS_CONFIG["rnet"]["hash"],  # noqa; E501
                cache_subdir=path.join("weights", "mtcnn"),
                hash_algorithm="sha256",
                extract=False,
            ),
        )
        load_weights(
            self.onet,
            utils.get_file(
                origin=WEIGHTS_CONFIG["onet"]["url"],  # noqa; E501
                file_hash=WEIGHTS_CONFIG["onet"]["hash"],  # noqa; E501
                cache_subdir=path.join("weights", "mtcnn"),
                hash_algorithm="sha256",
                extract=False,
            ),
        )

    def compute_inputs(self, images):
        raise NotImplementedError('MTCNN is not yet complete.')

    def compute_targets(self, collection):
        raise NotImplementedError('MTCNN is not yet complete.')

    def detect(
        self,
        image: Union[core.Image, np.ndarray],
        minsize=20,
        factor=0.709,
        threshold=0.8
    ) -> List[core.Annotation]:
        """Detects faces in an image, and returns bounding boxes and points for them.

        image: input image
        minsize: Minimum face size
        factor: The factor by which to step down scales in the
         image.
        threshold: The detection threshold.
        """
        image = image.view(core.Image)
        height, width = image.shape[:2]
        m = 12.0 / minsize  # Scale of MTCNN cell to minimum face size
        minl = min([height, width]) * m  # Image size to have a 12x12 region
        # correspond to minimum face size
        n_scales = math.ceil(math.log(12 / minl, factor))
        scales = [m * np.power(factor, n) for n in range(n_scales)]

        # Stage 1
        x1y1x2y2s = []
        for idx, scale in enumerate(scales):
            X = np.float32([
                image.resize(scale, interpolation=cv2.INTER_AREA).transpose(
                    1, 0, 2).scaled(-1, 1)
            ])
            y = [v.transpose(0, 2, 1, 3) for v in self.pnet.predict(X)]
            x1y1x2y2s_current = calculate_boxes(clf=y[0][0],
                                                reg=y[1][0],
                                                scale=scale,
                                                threshold=threshold,
                                                cell_size=12)
            if len(x1y1x2y2s_current) == 0:
                continue
            x1y1x2y2s.append(x1y1x2y2s_current[nms(x1y1x2y2s_current, 0.5,
                                                   "Union")])
        if len(x1y1x2y2s) == 0:
            return []
        x1y1x2y2s = np.concatenate(x1y1x2y2s)
        apply_regression(x1y1x2y2s)

        x1y1x2y2s = x1y1x2y2s[nms(x1y1x2y2s, 0.7, "Union")]

        # Stage 2
        pad_and_square(x1y1x2y2s, width=width, height=height, pad=0.1)
        x1y1x2y2s = run_stage(self.rnet,
                              image,
                              x1y1x2y2s,
                              threshold=threshold,
                              cell_size=24)
        if len(x1y1x2y2s) == 0:
            return []
        x1y1x2y2s = x1y1x2y2s[nms(x1y1x2y2s, 0.7, "Union")]

        # Stage 3
        pad_and_square(x1y1x2y2s, width=width, height=height, pad=0.1)
        x1y1x2y2s = run_stage(self.onet,
                              image,
                              x1y1x2y2s,
                              threshold=threshold,
                              cell_size=48)
        if len(x1y1x2y2s) == 0:
            return []
        x1y1x2y2s = x1y1x2y2s[nms(x1y1x2y2s, 0.5, "Min")]
        return [
            core.Annotation(
                selection=core.Selection(points=[[x1, y1], [x2, y2]]),
                category=self.annotation_config['face'],
                score=s
            ) for x1, y1, x2, y2, s in x1y1x2y2s[:, :5]
        ]



from abc import ABC, abstractmethod
from typing import List, Union, Tuple

from imgaug import augmenters as iaa
import numpy as np

from .. import metrics
from ..core import (
    SceneCollection,
    Annotation,
    Image
)


class Detector(ABC):
    def detect(
        self,
        image: Union[Image, np.ndarray],
        threshold: float=0.5
    ) -> List[Annotation]:
        """Run detection for a given image.

        Args:
            image: The image to run detection on

        Returns:
            A list of annotations
        """
        image = image.view(Image)
        image, scale = self._scale_to_model_size(image)
        X = self.compute_inputs([image])
        y = self.model.predict(X)
        annotations = self.invert_targets(
            y,
            images=[image],
            threshold=threshold
        )[0].annotations
        return [a.resize(1/scale) for a in annotations]

    def detect_batch(
        self,
        images: List[Union[Image, np.ndarray]],
        threshold: float=0.5,
        batch_size: int=32
    ) -> List[List[Annotation]]:
        """
        Perform object detection on a batch of images.

        Args:
            images: A list of images
            threshold: The detection threshold for the images
            batch_size: The batch size to use with the underlying model

        Returns:
            A list of lists of annotations.
        """
        images = [image.view(Image) for image in images]
        images, scales = list(
            zip(*[self._scale_to_model_size(image) for image in images])
        )
        X = self.compute_inputs(images)
        y = self.model.predict(X, batch_size=batch_size)
        scenes = self.invert_targets(
            y,
            images=images,
            threshold=threshold
        )
        return [
            [
                a.resize(1/scale) for a in scene.annotations
            ] for scene, scale in zip(scenes, scales)
        ]

    def _scale_to_model_size(self, image: Image):
        height, width = self.model.input_shape[1:3]
        if height is not None and width is not None:
            image, scale = image.fit(
                width=width,
                height=height
            )
        else:
            scale = 1
        return image, scale

    @abstractmethod
    def compute_inputs(self, images: List[Image]) -> np.ndarray:
        """Convert images into suitable model inputs. *You
        usually should not need this method*. For training,
        use `detector.train()`. For detection, use
        `detector.detect()`.

        Args:
            images: The images to convert

        Returns:
            The input to the model
        """
        pass

    @abstractmethod
    def compute_targets(
        self,
        collection: SceneCollection
    ) -> Union[List[np.ndarray], np.ndarray]:
        """Compute the expected outputs for a model. *You
        usually should not need this method*. For training,
        use `detector.train()`. For detection, use
        `detector.detect()`.

        Args:
            collection: The scene collection for which the
                outputs should be calculated.

        Returns:
            The output(s) as a list of numpy arrays
        """
        pass

    def freeze_backbone(self):
        """Freeze the body of the model, leaving the final classification and
        regression layer as trainable."""
        for l in self.backbone.layers:
            l.trainable = False
        self.compile()

    def unfreeze_backbone(self):
        """Unfreeze the body of the model, making all layers trainable."""
        for l in self.backbone.layers:
            l.trainable = True
        self.compile()

    def compute_Xy(self, collection: SceneCollection):
        return (
            self.compute_inputs(collection.images),
            self.compute_targets(collection)
        )

    def batch_generator(
        self,
        collection: SceneCollection,
        train_shape: Tuple[int, int, int],
        batch_size: int=1,
        augmenter: iaa.Augmenter=None
    ):
        while True:
            sample = collection.sample(
                n=batch_size
            ).augment(
                augmenter=augmenter
            ).fit(
                height=train_shape[0],
                width=train_shape[1]
            )[0]
            yield self.compute_Xy(sample)

    def train(
        self,
        training: SceneCollection,
        validation: SceneCollection=None,
        batch_size: int=1,
        augmenter: iaa.Augmenter=None,
        train_shape: Tuple[int, int, int]=None,
        **kwargs
    ):
        """Run training job. All additional keyword arguments
        passed to Keras' `fit_generator`. Note that, at a minimum,
        you must provide the `epochs` and `steps_per_epoch` arguments.
        If validation data is provided, you will also need to provide a
        `validation_steps` argument.

        Args:
            training: The collection of training images
            validation: The collection of validation images
            batch_size: The batch size to use for training
            augmenter: The augmenter for generating samples
        """
        assert training.annotation_config == self.annotation_config, \
            'The training set configuration clashes with detector'
        assert (
            validation is None or
            validation.annotation_config == self.annotation_config
        ), \
            'The validation set configuration clashes with detector'
        assert training.consistent, \
            'The training set has inconsistent annotation configuration'
        assert (
            validation is None or
            validation.consistent
        ), \
            'The validation set has inconsistent annotation configuration'
        if hasattr(self, 'training_model'):
            training_model = self.training_model
        else:
            training_model = self.model
        if train_shape is None:
            train_shape = training_model.input_shape[1:]
            assert all(s is not None for s in train_shape), \
                'train_shape must be provided for this model.'
        training_generator = self.batch_generator(
            collection=training,
            batch_size=batch_size,
            augmenter=augmenter,
            train_shape=train_shape
        )
        if validation is None:
            history = training_model.fit_generator(
                generator=training_generator,
                **kwargs
            )
        else:
            validation = validation.fit(
                height=train_shape[0],
                width=train_shape[1]
            )[0]
            history = training_model.fit_generator(
                generator=training_generator,
                validation_data=self.compute_Xy(validation),
                **kwargs
            )
        return history

    def mAP(self, collection: SceneCollection, iou_threshold=0.5):
        """Compute the mAP metric for a given collection
        of ground truth scenes.

        Args:
            collection: The collection to evaluate
            iou_threshold: The IoU threshold required for
                a match

        Returns:
            mAP score
        """
        pred = collection.assign(
            scenes=[
                s.assign(
                    annotations=self.detect(s.image, threshold=0)
                ) for s in collection
            ]
        )
        return metrics.mAP(
            true_collection=collection,
            pred_collection=pred,
            iou_threshold=iou_threshold
        )


from .preloaded import (  # noqa: F401
    load_voc2012,
    load_oxfordiiitpets,
    load_icdar2015,
    load_shapes
)
from .coco import load_coco  # noqa: F401
from .voc import load_voc  # noqa: F401
from .via import load_via  # noqa: F401